/* (c) 1999 P. Tuffery
 * laws.c
 * Everything concerning laws for detection of co-mutations
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include "global.h"
#include "tree.h"
#include "pevolve.h"
#include "pmodels.h"
#include "random.h"
#include "correlat.h"
#include "entropies.h"
#include "rates.h"           /* PT 2000 */

void doSIMDetect(TTree *tree, double scaling,  double deltaTreeLength, char *iRootFileName, double *targetRisk, int riskOptimization, int *maskDtct, int rfCommutMask, double dNumDataSets, int nCheckSims)
{
  extern int gDummy;
  extern double gRefTLen;

  double risk, lRisk;
  double deltaSign, deltaIncr;
  int lnPairs;
  double curTreeLength, refTreeLength, lnMaxPairs;
  int nTries;
  int j;
  int nPairs;
  int OK;
  int nTests;
  int setToss;
  int nTruePairs, dummyTruePairs;

  fprintf(stderr,"doSIMDetect:for risk %lf\n",*targetRisk);

  setToss = (riskOptimization) ? 0 : 1;
  
  /* Setup thresholds for basic risk ------------------------------------ */
  risk = *targetRisk;
  lRisk = risk;
  nTests = setLawsTrhld(&lRisk, maskDtct, rfCommutMask, dNumDataSets, setToss);

  /* Optionally optimize thresholds to fit at best given risk ----------- */
  if (riskOptimization) {
    if (1. - risk / (*targetRisk) >= 0.) {
      deltaSign = 1.;
    }
    if (1. - risk / (*targetRisk) < 0.) {
      deltaSign = -1.;
    }
    deltaIncr = 0.1;
    nTries = 0;
    
    do {
      if (fabs(1. - risk / (*targetRisk)) < 0.1) break;
      /*  	fprintf(stderr,"Target %lf, current %lf\n",*targetRisk,risk); */
      if (1. - risk / (*targetRisk) > 0.) {
	if (deltaSign < 0.) deltaIncr /= 1.5;
	deltaSign = 1.;
      }
      if (1. - risk / (*targetRisk) < 0.) {
	if (deltaSign > 0.) deltaIncr /= 1.5;
	deltaSign =-1.;
      }
      
      lRisk += deltaSign * deltaIncr * (*targetRisk);
      risk = lRisk;
      nTests = setLawsTrhld(&risk, maskDtct, rfCommutMask, dNumDataSets, setToss);
      ++nTries;
      if (++nTries > 200) break;
      /*  	  fprintf(stderr,"Try %d risk %lf\n",nTries,risk); */
    } while(1);
  }
  
  /* Perform detection ---------------------------------------------------- */
  /*  		  OutputThresholds("toto.trhld"); */
  nPairs = CommutDetectFromLaws(maskDtct, rfCommutMask, &nTruePairs, 2);
  
  /* Evaluate risk associated with thresholds ----------------------------- */
  refTreeLength = gRefTLen;
  if (gDummy) {
#if 0
    fprintf(stderr,"Dummy simulation for risk evaluation.\n");
    lnPairs = 0;
    for (j=0; j<nCheckSims; j++) {
      /* Dummy simulation (i.e. combinatorial exploration of repartition of mutations along tree */
      if (gDummy == 1) DummyEvolveSequences(tree, scaling, iRootFileName);
      else if (gDummy == 2) Dummy2EvolveSequences(tree, scaling, iRootFileName);
      else if (gDummy == 3) Dummy3EvolveSequences(tree, scaling, iRootFileName);
      DummyComutCount(tree, DcREFSEQ);
	  lnPairs += CommutDetectFromLaws(maskDtct, rfCommutMask, &dummyTruePairs, 0);
    }
    risk = (double) lnPairs / (1000. * (double) nTests); 
#else
    risk = *targetRisk;
#endif
  } else {
/*      fprintf(stderr,"Simulation for risk evaluation.\n"); */
    lnPairs = 0;
    for (j=0; j<nCheckSims; j++) {
      
      /* Implies an evoutionnary model behind */
      nTries = 0;
      do {
	OK = 1;
	EvolveSequences(tree, scaling, iRootFileName);
	curTreeLength = getTreeLength(tree, DcSIMSEQ);
	++nTries;
	if (++nTries > 200) {
	  fprintf(stderr,"Could not simulate correct tree length in %d tries\n",nTries);
	  exit(0);
	}
	/*  	    fprintf(stderr,"delta Tree length: %.2lf (ref %.2lf cur %.2lf)\n", */
	/*  		    fabs(curTreeLength - refTreeLength) / refTreeLength, */
	/*  		    curTreeLength, refTreeLength); */
	if (fabs(curTreeLength - refTreeLength) <= deltaTreeLength) {
	  /*  	      fprintf(stderr,"%d tries on length\n",nTries); */
	} else OK = 0;
#if 1
	if (OK) {
	  OK = checkRates(tree, DcSIMSEQ, refrates, gDeltaRate);
	}
#endif
      } while (!OK);
      
      TreeCMCount(tree, DcSIMSEQ);
      commutAsRfCommut();
      lnPairs += CommutDetectFromLaws(maskDtct, rfCommutMask, &dummyTruePairs, 0);
    }
    risk = (double) lnPairs / ((double) nCheckSims * (double) nTests); 
  }
  
  /* Max is nTests*risk + 1.64 * sqrt(nTests*risk*(1.-risk)) */
  /* and nTests*risk > 5 and nTests*(1.-risk) > 5 */
  lnMaxPairs = *targetRisk * (double) nTests + 1.64 * sqrt(*targetRisk * (double) nTests * (1. - *targetRisk));
  fprintf(stdout,"For risk %.4lf (wanted %.4lf) : %d pairs detected (%d true) (%d tests, expct is %.1lf, max acceptable is %.1lf)\n",
	  risk,*targetRisk,nPairs,nTruePairs, nTests,*targetRisk * (double) nTests,lnMaxPairs);
  *targetRisk = risk;
}


