/*
 * PSeq-Gen
 * (c) 1997 N Grassly and A Rambaut
 *	Department of Zoology
 *	Oxford University
 *	pseq-gen.c
 *
 * CMPSeq-Gen addon by
 * P. Tuffery (1997-2000) 
 * INSERM U436, Paris, France
 *
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <math.h>

#ifdef __MWERKS__
#include <console.h>
#endif

#include "global.h"
#include "treefile.h"
#include "pevolve.h"
#include "pmodels.h"
#include "CSFuncs.h"
#include "Random.h"

/* =======================================================
 * A *very* basic procedure to optimize tree scaling so that 
 * simulation is centered on reconstructed phylogeny.
 * =======================================================
 */
void scalingEquilibration(TTree *tree, char *iRootFileName, 
			  int nCycle, int nSim, 
			  double scaling, double *scFac, 
			  double refTLen, double deltaTLen,
			  double aEpsi)
{
  int aCycle;
  int aIter;
  double curTLen, meanTLen;

  fprintf(stderr,"treeScalingEquilibration: Will perform %d cycles\n",nSim);

  for (aCycle=0; aCycle<nCycle; aCycle++) {

    meanTLen = 0.;

    for (aIter=0; aIter<nSim; aIter++) {
      
      doEvolveSequences(tree, iRootFileName, tree->pCs->p,
			scaling*(*scFac),
			1000., 20,
			&curTLen, deltaTLen, 10000., 20);

      meanTLen += curTLen;
    }
    meanTLen /= (double) nSim;
    
    if (fabs(meanTLen - refTLen) / refTLen < aEpsi) {
      fprintf(stderr,"Iter %d: scaling factor %lf (delta %lf)\n",
	      aCycle,
	      *scFac,
	      fabs(meanTLen - refTLen) / refTLen);
      break;
    } else {
      *scFac /= meanTLen / refTLen;
      fprintf(stderr,"Cycle %d: scaling factor %lf (treeLength Obs : %lf Ref : %lf\n",
	      aCycle,*scFac,
	      meanTLen, refTLen);
    }
  }
}


