/* This program is based on PSeq-Gen
 * (c) 1997 N Grassly and A Rambaut
 *	Department of Zoology
 *	Oxford University
 *	pseq-gen.c
 * 
 * CS-Pseq-Gen:
 * by P. Tuffery (1997-2001) for the implementation
 * of data constraints and the simulation / detection of co-substitutions 
 * P. Tuffery (1997-2000) 
 * INSERM U436, Universite Paris 7, France
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <math.h>

#ifdef __MWERKS__
#include <console.h>
#endif

#include "global.h"
#include "treefile.h"
#include "tree.h"
#include "pevolve.h"
#include "pmodels.h"
#include "progress.h"
#include "random.h"

#include "CSTypes.h"
#include "Poisson.h"
#include "Random.h"
#include "Rates.h"  
#include "RefSeq.h"  
#include "Root.h"  
#include "TreeUtils.h"  


#define PROGRAM_NAME "cspseq-gen"
#define VERSION_NUMBER "Version 1.0"


extern int gNTries, gNEvolve;

int gDoPrint = 0;
int gOutTrueCM = 0;
int gPrfCorrp = 0;
int gMaxClstrSze = 2;
double gMaxClstrDiffRates = 0.05;
int gSilentGap = 0;
int nClstr = 0;

int gOutNodes = 0;

/* Various variables related to equilibration procedures */
int gEquilSttgy = 0;
int gNScEquilCycles = 50; 
int gNScEquilSim = 50;
int gNRateEquilCycles = 50;
int gNRetesEquilSim = 50;

/* Detection of co-substituting pairs */
int CS_detect = 0;

int gOptimizeRisk = 1;
int gNCMRisks = 3;
double sumCMTrhld[3];
double sumCMRisk[3];
double CMTrhld[3] = {0.05,0.01,0.001};
int numCMDatasets = 50;
int gNRisks = 3;
double gRisks[3] = {0.05,0.01,0.001};

int    gSCS = 0;       /* trigger the constrained simulations */
double gEDTL = 1000.;  /* Default values, i.e. no control     */
double gEDBL = 1000.;   
double gRiskDeltaTreeLength   = 0.05; /* Default risk values when control ON */
double gRiskDeltaBranchLength = 0.05; 
double gDeltaTLen = 0.3;

int gNBMaxTries = 50;  /* Max number of tries per branch */
int gNTMaxTries = 50;  /* Max number of tries per tree   */

int gUseRefRates = 0;
int gUseRefBLen  = 0;

extern double gMinRate;
extern double gMaxRate;
double gMinCorrRate = 0.2;
double gMaxCorrRate = 20.;
int DEBUG = 0;

int gDoSIMStat    = 1;
int gDoCMStat     = 0;
int gDoHBStat     = 0;
int gRfCommutMask = 0;
int gNoCtctDtct   = 0;

int gDoEquilibration    = 0;
int gDoEquilTreeLength  = 0;
int gNEquilSim          = 50;
int gNumEqTreeCycles    = 50;
int gNumTreeEqDatasets  = 200;
int gDoEquilRates       = 0;
int gNumRatesEqCycles   = 5;
int gNumRatesEqDatasets =  200;

char *ModelNames[4]={
	"PAM",
	"JTT",
	"MTREV",
	"GENERAL"
};

char *MutationName[numMutation]={     /* MD 97 */
	"SIM",           
	"COR",          
	"BOT"         
};                      
int Fasta=0;
int gShowHomol= 0;
int averageNumSubDist = 0;
#if 0
int gEvolveMode;
int gBranchMutMode;
int myseed=-1;
#endif

int *mask     = NULL;
int *maskDtct = NULL;

int gNoGapMask = 1;

/* int gLengthMode = DcLENGTHSFROMOBS; */
int gLengthMode = DcLENGTHSFROMDIST;

int printHelp, treeFile, numDatasets, numTrees;
int scaleTrees, scaleBranches, mfreqSet;
double treeScale, branchScale;
char iTreeFileName[BUFSIZ];
char iCfgFileName[BUFSIZ];
char oCfgFileName[BUFSIZ];
char iRatesFileName[BUFSIZ];
char iRootFileName[BUFSIZ];
char iRefSeqFileName[BUFSIZ];
char iPAMLRefSeqFileName[BUFSIZ];
char iRefNamesFileName[BUFSIZ];
char iPAMLRefNamesFileName[BUFSIZ];
char iLawsFileName[BUFSIZ];
char iMaskFileName[BUFSIZ];
char iConvFileName[BUFSIZ];
char oLawsFileName[BUFSIZ];
char oSeqFileName[BUFSIZ];
char oBoolFileName[BUFSIZ];
char oLisFileName[BUFSIZ];

FILE *oLisFich = NULL;
FILE *oSeqFich = NULL;

/* prototypes */

static void PrintTitle();
static void PrintUsage();

/* functions */

static void PrintTitle(FILE *f)
{
	fprintf(f, "\nConstrained Protein Sequence Generator - %s\n", PROGRAM_NAME);
	fprintf(f, "%s\n\n", VERSION_NUMBER);
	fprintf(f, "P. Tuffery, Equipe de Bioinformatique Genomique et Moleculaire\n");
	fprintf(f, "INSERM U436, Paris, France.\n\n");
	fprintf(f, "An extension of pseq-gen.\n");
	fprintf(f, "(c) Copyright, 1997 Nick Grassly and Andrew Rambaut\n");
	fprintf(f, "Department of Zoology, University of Oxford\n");
	fprintf(f, "South Parks Road, Oxford OX1 3PS, U.K.\n\n");
}

static void PrintParams(FILE *f)
{ 
  if (iTreeFileName[0] != '\000') 
    fprintf(stderr, "Tree file             : %s\n", iTreeFileName);
  if (iCfgFileName[0] != '\000') 
    fprintf(stderr, "Cfg file              : %s\n", iCfgFileName);
  if (oCfgFileName[0] != '\000') 
    fprintf(f, "Cfg file  (output)         : %s\n", oCfgFileName);
  if (iRatesFileName[0] != '\000') 
    fprintf(stderr, "Site varibility file  : %s\n", iRatesFileName);
  if (iRootFileName[0] != '\000') 
    fprintf(stderr, "Root sequence file    : %s\n", iRootFileName);
  if (iRefNamesFileName[0] != '\000') 
    fprintf(stderr, "RefNames file         : %s\n", iRefNamesFileName);
  if (iPAMLRefNamesFileName[0] != '\000') 
    fprintf(stderr, "PAMLRefNames file     : %s\n", iPAMLRefNamesFileName);
  if (iRefSeqFileName[0] != '\000') 
    fprintf(stderr, "Ref sequences file    : %s\n", iRefSeqFileName);
  if (iPAMLRefSeqFileName[0] != '\000') 
    fprintf(stderr, "PAMLRef sequences file: %s\n", iPAMLRefSeqFileName);
  if (iMaskFileName[0] != '\000') 
    fprintf(stderr, "Mask file             : %s\n", iMaskFileName);
  if (oSeqFileName[0] != '\000') 
    fprintf(stderr, "Out sequences file    : %s\n", oSeqFileName);
  if (iConvFileName[0] != '\000') 
    fprintf(stderr, "In conversion file    : %s\n", iConvFileName);
}

/* =====================================================================
 * Help about command line arguments
 * =====================================================================
 */
static void argstrDeltaPrmtrsUsage()
{
  fprintf(stderr,"Constraining simulations around reference data.\n");
  fprintf(stderr,"  -sCS : enable constraints with default values.  (type I error 0.05 for dT and dB, 0.05 for dR)\n");
  fprintf(stderr,"  -dT <n stdev> (0.05): authorized tree lenght variation (1000. for no constaint)\n");
  fprintf(stderr,"  -dB <risk> (0.05): authorized risk on branch lenght variation (1000. for no constaint)\n\n");
  fprintf(stderr,"  -dR <0.5,0.1,0.01,0.>(0.01): deltaRates allowed (0. for no constaint) (only if reference rates are specified)\n");
  fprintf(stderr,"  -nBt <x> (50): max tries to get simulation satisfying constraints\n");
  fprintf(stderr,"  -nTt <x> (50): max tries to get simulation satisfying constraints\n");
  fprintf(stderr,"  -doTLOpt : attempt to automatically adjust scaling to center simulated tree length on reference tree lenght\n\n");
}

/* =====================================================================
 * Help about command line arguments
 * =====================================================================
 */
static void argstrMiscPrmtrsUsage()
{
  fprintf(stderr,"Miscellaneaous\n");
  fprintf(stderr,"  -oseq <fname> : redirect sequence output to file\n");
  fprintf(stderr,"  -onodes       : output intermediate node sequences\n");
  fprintf(stderr,"  -nosv         : do not output simulated sequences\n\n");
}

static void PrintUsage()
{ 
	fprintf(stderr, "Usage: seq-gen [-m MODEL] [-l #] [-n #] [-s # | -d #] [-g #]\n               [-f #ala,arg...] [-p] [-q] [-h] [treefile]\n");
	fprintf(stderr, "  -m: MODEL = PAM, JTT, MTREV [default = PAM]\n");
	fprintf(stderr, "  -l: # = sequence length [default = 1000].\n");
	fprintf(stderr, "  -n: # = simulated datasets per tree [default = 1].\n");
	fprintf(stderr, "  -s: # = branch length scaling factor [default = 1.0].\n");
	fprintf(stderr, "  -d: # = total tree scale [default = use branch lengths].\n");
	fprintf(stderr, "  -g: # = shape for gamma rate heterogeneity [default = none].\n");
	fprintf(stderr, "  -f: # = amino acid frequencies e=equal [default = matrix freqs].\n");
	fprintf(stderr, "  -p: Use the relaxed PHYLIP format\n");
	fprintf(stderr, "  -h: Give this help message\n");
	fprintf(stderr, "  -q: Quiet\n");
	fprintf(stderr, "  -v: Verbose\n");
	fprintf(stderr, "  treefile: name of tree file [default = trees on stdin]\n\n");
}


static void ReadParams(int argc, char **argv)
{
  int i, j, estimate;
  char ch, *P, st[3], mst[3];
  
  model=PAM;
  gEvolveMode=SIM; 
  gBranchMutMode = DcCORRNOTALLOW;
  
  scaleTrees=0;
  treeScale=0.0;
  scaleBranches=0;
  branchScale=1.0;
  mfreqSet=1;
  rateHetero=NoRates;
  gammaShape=1.0;
  
  numBases=1000;
  numDatasets=2;
  
  verbose=0;
  relaxedPhylip=0;
  quiet=0;
  
  treeFile=0;

  pR = InitRates(numBases);

  for (i=1; i<argc; i++) {
    P=argv[i];
    
    if (*P!='-') {
      if (treeFile) {
	fprintf(stderr, "Illegal command parameter: %s\n\n", argv[i]);
	PrintUsage();
	exit(0);
      }
      treeFile=1;
      strcpy(iTreeFileName, argv[i]);
      continue;
    } else if (!strcmp(argv[i],"-dbg")) {
      DEBUG = 1;
      continue;
    } else if (!strcmp(argv[i],"-itree")) {
      treeFile=1;
      strcpy(iTreeFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-icnv")) {
      strcpy(iConvFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-iRoot")) {
      strcpy(iRootFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-iMask")) {
      strcpy(iMaskFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-iRefNames")) {
      strcpy(iRefNamesFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-iPAMLRefNames")) {
      strcpy(iPAMLRefNamesFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-refSeqBLen")) {
      gUseRefBLen = 1;
      continue;
    } else if (!strcmp(argv[i],"-ilaws")) {
      strcpy(iLawsFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-olaws")) {
      strcpy(oLawsFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-oBool")) {
      strcpy(oBoolFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-iRefSeq")) {
      strcpy(iRefSeqFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-iPAMLRefSeq")) {
      strcpy(iPAMLRefSeqFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-oseq")) {
      strcpy(oSeqFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-onodes")) {
      gOutNodes = 1;
      continue;
    } else if (!strcmp(argv[i],"-nosv")) {
      oSeqFich = NULL;
      continue;
    } else if (!strcmp(argv[i],"-CSDetect")) {
      CS_detect = 1;
      continue;
    } else if (!strcmp(argv[i],"-noCtctDtct")) {
      gNoCtctDtct = 1;
      continue;
    } else if (!strcmp(argv[i],"-allowGap")) {
      gNoGapMask = 0;
      continue;
    } else if (!strcmp(argv[i],"-risk")) {
      gNRisks = 1;
      gRisks[0] = atof(argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-MHMAACls")) {
      setMHMAAClassFilter();
      continue;
    } else if (!strcmp(argv[i],"-HBAACls")) {
      setHBAAClassFilter();
      continue;
    } else if (!strcmp(argv[i],"-d1AACls")) {
      setDummy1MHMAAClassFilter();
      continue;
    } else if (!strcmp(argv[i],"-d2AACls")) {
      setDummy2MHMAAClassFilter();
      continue;
    } else if (!strcmp(argv[i],"-PolarAACls")) {
      setPolarAAClassFilter();
      continue;
    } else if (!strcmp(argv[i],"-stdAACls")) {
      setAAClassFilter();
      continue;
    } else if (!strcmp(argv[i],"-noNullRf")) {
      gRfCommutMask = 1;
      continue;
    } else if (!strcmp(argv[i],"-o")) {
      strcpy(oLisFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-doCM")) {
      gDoCMStat = 1;
      continue;
    } else if (!strcmp(argv[i],"-doHB")) {
      gDoHBStat = 1;
      continue;
    } else if (!strcmp(argv[i],"-noRiskOpt")) {
      gOptimizeRisk = 0;
      continue;
    } else if (!strcmp(argv[i],"-doRootEquil")) {
      gDoEquilibration = 1;
      continue;
    } else if (!strcmp(argv[i],"-nEquil")) {
      gNEquilSim = atoi(argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-doTLOpt")) {
      gDoEquilTreeLength = 1;
      continue;
    } else if (!strcmp(argv[i],"-outCM")) {
      gOutTrueCM = 1;
      continue;
    } else if (!strcmp(argv[i],"-trueDist")) {
      gLengthMode = DcLENGTHSFROMDIST;
      continue;
    } else if (!strcmp(argv[i],"-treeDist")) {
      gLengthMode = DcLENGTHSFROMTREEFILE;
      continue;

    } else if (!strcmp(argv[i],"-sCS")) { /* constrain simulations */
      gSCS = 1;
    } else if (!strcmp(argv[i],"-dT")) {  /* Risk on delta tree length   */
      gRiskDeltaTreeLength = atof(argv[i+1]);
      if (whichEpsilonValue(gRiskDeltaTreeLength) == -1.) {
	fprintf(stderr,"Sorry: -dT %lf: not tabulated risk\n", gRiskDeltaTreeLength);
	listEpsilonValues();
	exit(0);
/*  	fprintf(stderr,"       reverting to 0.05\n"); */
/*  	gRiskDeltaTreeLength = 0.05; */
      }
      i++;
      continue;
    } else if (!strcmp(argv[i],"-seed")) {  /* random generator seed */
      gSeed = atol(argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-dR")) {  /* Epsilon on delta tree length   */
      gDeltaRate = atof(argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-dB")) {  /* Risk on delta branch length */
      gRiskDeltaBranchLength = atof(argv[i+1]);
      if (whichEpsilonValue(gRiskDeltaBranchLength) == -1.) {
	fprintf(stderr,"Sorry: -dB %lf: not tabulated risk\n", gRiskDeltaBranchLength);
	listEpsilonValues();
	exit(0);
/*  	fprintf(stderr,"       reverting to 0.05\n"); */
/*  	gRiskDeltaBranchLength = 0.05; */
      }
      i++;
      continue;
    } else if (!strcmp(argv[i],"-nBt")) {  /* nTries / branch */
      gNBMaxTries = atoi(argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-nTt")) {  /* nTries / branch */
      gNTMaxTries = atoi(argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-nCMSteps")) {
      numCMDatasets = atoi(argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-minRate")) {
      gMinRate = atof(argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-maxRate")) {
      gMaxRate = atof(argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-bScale")) {
      scaleBranches=1;
      branchScale = atof(argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-sGap")) {
      gSilentGap = 1;
      i++;
      continue;

      /* CS Params duplicated HERE */
    } else if (!strcmp(argv[i],"-icfg")) {
      strcpy(iCSCfgFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-ocfg")) {
      strcpy(oCSCfgFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-tFac")) {
      CSClsParms.tFac = atof(argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-rndCScl")) {
      CSClsParms.rndDefine = 1;
      continue;
    } else if (!strcmp(argv[i],"-nCScl")) {
      CSClsParms.nCls = atoi(argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-CSclMaxSze")) {
      CSClsParms.maxSze = atoi(argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-CSclMinSze")) {
      CSClsParms.minSze = atoi(argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-CSclMinR")) {
      CSClsParms.minRate = atof(argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-CSclMaxR")) {
      CSClsParms.maxRate = atof(argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-CSclDiffR")) {
      CSClsParms.maxDiffRate = atof(argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-CS")) {
      gCSUse = 1;
      continue;

      /* Rates Params duplicated HERE */
    } else if (!strcmp(argv[i],"-iRates")) {
      rateHetero=UserRates;
      strcpy(iRatesFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-refRates")) {
      gUseRefRates = 1;
      continue;
    } else if (!strcmp(argv[i],"-doRatesOpt")) {
      pR->doEquil = 1;
      continue;

      /* Back to standard pseq-gen parameters */
    } else {
      while (*P) {
	*P=toupper(*P);
	P++;
      }
      P=argv[i];
      P++;
      ch=*P;
      P++;
      switch (ch) {
      case 'H':
	printHelp=1;
	break;
      case 'M':
	if (GetStrParam(argc, argv, &i, P, st, 3)) {
	  fprintf(stderr, "Bad (or missing) Model Code: %s\n\n", argv[i]);
	  exit(0);
	}
	      
	P=st;
	model=-1;
	for (j=PAM; j<GENERAL; j++) {
	  if (strncmp(P, ModelNames[j], 3)==0)
	    model=j;
	}
	if (model==-1) {
	  fprintf(stderr, "Unknown Model: %s\n\n", argv[i]);
	  exit(0);
	}
	      
	break;
      case 'L':
	if (GetIntParams(argc, argv, &i, P, 1, &numBases) || numBases<1) {
	  fprintf(stderr, "Bad (or missing) sequence length: %s\n\n", argv[i]);
	  exit(0);
	}
	break;
      case 'N':
	if (GetIntParams(argc, argv, &i, P, 1, &numDatasets) || numDatasets<1) {
	  fprintf(stderr, "Bad (or missing) number of datasets: %s\n\n", argv[i]);
	  exit(0);
	}
	break;
      case 'G':
	rateHetero=GammaRates;
	if (GetDoubleParams(argc, argv, &i, P, 1, &gammaShape)) {
	  fprintf(stderr, "Bad Gamma Shape: %s\n\n", argv[i]);
	  exit(0);
	}
	break;
      case 'F':
	mfreqSet=0;
	if (*P=='E'){
	  for(i=0;i<NUM_AA;i++)
	    freq[i]=0.05;
	}else{
	  if (GetDoubleParams(argc, argv, &i, P, NUM_AA, freq)) {
	    fprintf(stderr, "Bad Amino Acid Frequencies: %s\n\n", argv[i]);
	    exit(0);
	  }
	}
	break;
      case 'D':
	scaleTrees=1;
	if (GetDoubleParams(argc, argv, &i, P, 1, &treeScale) || treeScale<=0.0) {
	  fprintf(stderr, "Bad Total Tree Scale: %s\n\n", argv[i]);
	  exit(0);
	}
	if (scaleBranches) {
	  fprintf(stderr, "You can't specify both the -d and -s options\n\n", argv[i]);
	  exit(0);
	}
	break;
      case 'S':
	scaleBranches=1;
	if (GetDoubleParams(argc, argv, &i, P, 1, &branchScale) || branchScale<=0.0) {
	  fprintf(stderr, "Bad Branch Length Scale: %s\n\n", argv[i]);
	  exit(0);
	}
	if (scaleTrees) {
	  fprintf(stderr, "You can't specify both the -d and -s options\n\n", argv[i]);
	  exit(0);
	}
	break;
      case 'P':
	relaxedPhylip=1;
	break;
      case 'Q':
	quiet=1;
	break;
      case 'V':
	verbose=1;
	break;
      case 'W':
	gShowHomol=1;
	break;         
      case 'X':      
	relaxedPhylip=1;
	Fasta=1;   
	break;            
      case 'Y':
	averageNumSubDist=1;
	break; 
				/* MD 97 */
      case 'Z':  
	if (GetStrParam(argc, argv, &i, P, mst, 3)) {
	  fprintf(stderr, "Bad (or missing) Mutation Code: %s\n\n", argv[i]);
	  exit(0);	        
	}                        
	P=mst;		        
	gEvolveMode=-1;            
	for (j=SIM; j<numMutation; j++) {  
	  if (strncmp(P, MutationName[j], 3)==0) 
	    gEvolveMode=j;                    
	}	 			  
	if (gEvolveMode > SIM) {
	  gBranchMutMode = DcCORRALLOW;
	  gCSUse = 1;
	}
	if (gEvolveMode==-1) {               
	  fprintf(stderr, "Unknown Mutation: %s\n\n", argv[i]);
	  exit(0); 
	}          
	if (DEBUG) fprintf(stderr,"Initial default evolve mode is %d (%s / %s%d %s%d %s%d)\n",
			   gEvolveMode,P,
			   MutationName[0],SIM,MutationName[1],COR,MutationName[2],BOT);
	break;            
	      
      default:
	fprintf(stderr, "Illegal command parameter: %s\n\n", argv[i]);
	PrintUsage();
	exit(0);
	break;
      }
    }
  }
	
}

static void PrintVerbose(FILE *f)
{
int i;
	fprintf(f, "Simulating %d taxa, %d bases\n", numTaxa, numBases);
	fprintf(f, "  for %d tree(s) with %d dataset(s) per tree\n\n", numTrees, numDatasets);
	
	if (scaleTrees) {
		fprintf(f, "Branch lengths of trees scaled so that tree is %G from root to tip\n\n", treeScale);
	} else if (scaleBranches) {
		fprintf(f, "Branch lengths of trees multiplied by %G\n\n", branchScale);
	} else {
		fprintf(f, "Branch lengths assumed to be number of substitutions per site\n");
	}
	fprintf(f, "Substitution model is %s\n\n", ModelNames[model]);
	if (rateHetero==GammaRates) {
		fprintf(f, "Discrete gamma rate heterogeneity:\n");
		fprintf(f, "    shape = %lf\n", gammaShape);
	} else
		fprintf(f, "Rate homogeneity of sites.\n");
	fprintf(f, "Amino Acid Frequencies\n");
	for(i=0;i<NUM_AA;i++)
		fprintf(f, "%c: %lf\t", aminoAcids[i], freq[i]);
	fprintf(f, "\n\n");
	/*  fprintf(f, "Evolution mode: %s\n\n", MutationName[gEvolveMode]);     */     /* MD 97 */
	fflush(f);
	if(verbose){
		printEigen(f);
		printPij(f);
		fflush(f);
	}
}

/* ==============================================================
 * HERE WE GO !
 * ==============================================================
 */
int main(int argc, char **argv)
{
  int aSim;
  int nCS = 0;

  int i, j, treeNo, tmpI;
  FILE *tree_fv;
  TTree *tree;
  clock_t totalStart;
  double totalSecs, scale;
  char lbl1[80], lbl2[80];

  int lnPairs;
  double risk;
  double dNumDataSets;
  double lnMaxPairs;
  int nPairs, nTests, nIter;
  double gDeltaSign, gDeltaIncr, lRisk;
  int nTries;

  double curTLen;
  double meanTLen,stdvTLen;
  double meanCs, stdvCs;

  int *dummyMat = NULL, *rstab = NULL;
  int OK;
  int curTip;
  char aC;
  int nRatesFail = 0;
  int nTruePairs;
  
  int refSeqp = 0;

  double meanCmt, varCmt;
#ifdef __MWERKS__
  argc=ccommand(&argv);
#endif

  /* Some initialisations ------------------------------ */
  if (setvbuf( stderr, NULL, _IOLBF, 512 ) ||
      setvbuf( stdout, NULL, _IOLBF, 512 ) ) {
    fprintf(stderr, "Failed to buffer stdout and stderr\n");
    exit(0);
  }
 	
  totalStart = clock();

  iCfgFileName[0]        = '\000';  
  oCfgFileName[0]        = '\000';  
  oSeqFileName[0]        = '\000';
  oLisFileName[0]        = '\000';
  iRatesFileName[0]      = '\000';
  iRootFileName[0]       = '\000';
  iRefNamesFileName[0]   = '\000';
  iPAMLRefNamesFileName[0] = '\000';
  iRefSeqFileName[0]     = '\000';
  iPAMLRefSeqFileName[0] = '\000';
  iLawsFileName[0]       = '\000';
  iMaskFileName[0]       = '\000';
  oLawsFileName[0]       = '\000';
  iTreeFileName[0]       = '\000';
  iConvFileName[0]       = '\000';
  oBoolFileName[0]       = '\000';
  CS_CSClsParmsDefaults(&CSClsParms);

  oSeqFich = stdout;
  
  if (DEBUG) fprintf(stderr,"Will read command line ...\n");
  ReadParams(argc, argv);

#if 1 /* Not used since ReadParams hardly modifiable, but to be left
       for internal inits */
  /* --> Addon for parsing CS related command line options */
  parseArgstrCSPrmtrs(argc,argv);
  /* --> Addon for parsing Rates related command line options */
  parseArgstrRatesPrmtrs(argc,argv);
#endif

  if (gSCS) {    /* setup the constrained simulations values */
    gEDTL = gRiskDeltaTreeLength;
    gEDBL = gRiskDeltaBranchLength;
  } 


  if (printHelp) {
    PrintTitle(stderr);
    PrintUsage();
    argstrRootPrmtrsUsage();
    argstrRefSeqPrmtrsUsage();
    argstrRatesPrmtrsUsage();
    argstrDeltaPrmtrsUsage();
    argstrCSPrmtrsUsage();
    argstrMiscPrmtrsUsage();
    exit(0);
  }

  oLisFich = stderr;
  if (oLisFileName[0] != '\000') {
    oLisFich = fopen(oLisFileName,"w");
    if (oLisFich == NULL) oLisFich = stderr;
  }
  
  if (!quiet) {
    PrintTitle(oLisFich);
    PrintParams(oLisFich);
    CSParams(oLisFich);
    RatesParams(oLisFich);
  }
  
  pAAClassInfo();

  /* We use drand48() instead of original PSeq-Gen generator */
  if (gSeed == -1) gSeed=time(NULL);  
  fprintf(oLisFich,"Pseudo-random generator seed %ld\n\n",gSeed);
  SetSeed(gSeed, oLisFich);
  initRnd(&gSeed);

  if (DEBUG) fprintf(stderr,"Will set the model (%d)...\n", model);
  SetModel(model);

  numTaxa=-1;
  scale=1000.0;/* therefore bl=1 => 1 change in hundred */
	
  if ((tree=NewTree())==NULL) {
    fprintf(stderr, "Out of memory\n");
    exit(0);
  }
	
  if (treeFile) {
    if ( (tree_fv=fopen(iTreeFileName, "r"))==NULL ) {
      fprintf(stderr, "Error opening tree file: '%s'\n", iTreeFileName);
      exit(0);
    }
  } else 
    tree_fv=stdin;
  
  numTrees=CountTrees(tree_fv);
  

  /* Add-on ============================================= */
  tree->nSites   = numBases;
  tree->nESites  = numBases;
  tree->pCs      = NULL;
  tree->siteMode = NULL;
  tree->siteMode = calloc(sizeof(int) *  tree->nSites, 1);
  for (i=0; i<tree->nSites; i++) {
    tree->siteMode[i] = DcEVOLVING;
  }
  tree->rates = NULL;
  tree->rates = calloc(sizeof(double) *  tree->nSites, 1);
  for (i=0; i<tree->nSites; i++) {
    tree->rates[i] = 1.;
  }

  tree->sumS     = NULL;
  tree->sumS2    = NULL;
  tree->aDVec    = NULL;

  tree->distMode = DcLENGTHSFROMOBS;
  tree->minRate  = gMinRate;
  tree->maxRate  = gMaxRate;

  initTreeCount(tree);
  
  if (oSeqFileName[0] != '\000') {
    oSeqFich = fopen(oSeqFileName,"w");
  }
  
  /* ==========================================
   * Start simulation for tree
   * ==========================================
   */
  if (DEBUG) fprintf(stderr,"Will simulate ...\n");
  fflush(stderr);
  treeNo=0;
  do {
    if (DEBUG) fprintf(stderr,"Will read tree file ...\n");
    ReadTree(tree_fv, tree, treeNo+1);

    /* --> Addon for the root sequence randomly generated is default */
    tree->rootMode = DcRANDOMROOT;
    if (iRootFileName[0] != '\000') tree->rootMode = DcUSERROOT;
    tree->refSeqp  = DcREFSEQNOUSE;
    WriteTree(stderr, tree);
    tree->refTLen  = 0.;
    SumTreeDist(tree, tree->root, &tree->refTLen);
    fprintf(stderr,"Tree Total Lenght %lf\n", tree->refTLen);

  /*    SetLengthsAsTrueLengths(tree, tree->root, DcSIMSEQ, gLengthMode); */

    /* --> Addon for the labels of the internal nodes (required since
     * we would like to input the externally re-constructed sequences 
     * Additionnaly, one can use such facility to trigger on/off
     * co-substitutions along specific branches.
     */
    if ((iRefNamesFileName[0] != '\0') || (iPAMLRefNamesFileName[0] != '\0')) {
      if (iRefNamesFileName[0] != '\0') inputRefNames(tree, iRefNamesFileName, oLisFich);
      if (iPAMLRefNamesFileName[0] != '\0') inputPAMLRefNames(tree, iPAMLRefNamesFileName, oLisFich);
      DescribeTree(stdout,tree,tree->root);

      /* Extra facility to root tree on given branch (uses node labels) */
      /* We need rooted tree */
      if (!tree->rooted) {
	printf("Enter branch for root: ");
	fscanf(stdin,"%s%s",lbl1,lbl2);
	rootUTreeAtLabels(oLisFich, tree,lbl1,lbl2,0.);
      }
      fprintf(stderr,"Rooted tree ...\n");
    }
    
    if (treeNo==0) {
      numTaxa=tree->numTips;
      CreateSequences(tree, numBases);
      
      /* --> Addon for rate heterogeneity */
      /* if (rateHetero != NoRates) SetRates(numBases, gammaShape); */

      if (scaleTrees) { 
	if (!tree->rooted) {
	  fprintf(stderr, "To scale tree length, they must be rooted and ultrametric.\n");
	  exit(0);
	}
	scale*=(treeScale/tree->totalLength);
      } 
      /* else if (scaleBranches) */
      /* scale*=branchScale; */
      
    } else if (tree->numTips!=numTaxa) {
      fprintf(stderr, "All trees must have the same number of tips.\n");
      exit(0);
    }
    
    /* --> Addon for input of the root sequence 
     * We may need it since we have a reconstructed phylogeny, and since
     * mutation rates depend on the nature of the amino acids
     */
    if (iRootFileName[0] != '\0') {
      if (DEBUG) fprintf(stderr,"Will input root sequence (%s)...\n",iRootFileName);
      InputRootSequence(iRootFileName, stderr,tree);
      if (DEBUG) fprintf(stderr,"Root sequence input done ...\n");
    }

    /* --> Addon for input of the mutation rates per site
     * read rates from disk if asked for.
     */
    if (DEBUG) fprintf(stderr,"Will input rates ...\n");
    InputRates(iRatesFileName);

    /* --> Addon for initialization of CS fields */
    tree->pCs = CS_Setup(tree, iCSCfgFileName, tree->pCs, numBases, oLisFich);
    tree->pCs->active = gCSUse;
    if (DEBUG) fprintf(stderr,"Will init co-subsitution detection fields ...\n");
/*      initCMDetect(); */
    CS_detectInit(tree->pCs);

    /* --> Addon for input of the reference sequence 
     * We may need it since we need a count of co-subsitutions 
     * in reconstructed phylogeny
     * We may also need it to parameterize the simulation to fit
     * data at best.
     */
    if ((iRefSeqFileName[0] != '\0') || (iPAMLRefSeqFileName[0] != '\0')) {
      refSeqp = 1;

      /* Input reference sequences (inferred ancestral sequences) */
      if (iRefSeqFileName[0] != '\0') RefSeqUse(tree, iRefSeqFileName, 0, 
						gNoGapMask, 
						gUseRefRates, gUseRefBLen,
						&tree->refTLen);

      if (iPAMLRefSeqFileName[0] != '\0') RefSeqUse(tree, 
						    iPAMLRefSeqFileName, 1,
						    gNoGapMask, 
						    gUseRefRates, gUseRefBLen,
						    &tree->refTLen);


      if (DEBUG) fprintf(stderr,"Reference sequences usage done ...\n");
    } /* iRefSeqFileName */



    /* --> Addons for co-evolving sites */
    tree->pCs = CS_Setup2(tree, iCSCfgFileName, tree->pCs, numBases, oLisFich);
    if (oCSCfgFileName[0] != '\000')
      CS_CfgWrite(oCSCfgFileName, tree->pCs);

    if (iRootFileName[0] != '\0') {  /* Check gapped sites to invalidate them. */
      if (DEBUG) fprintf(stderr,"Will scan root sequence (%s)...\n",iRootFileName);
      ScanRootSequence(iRootFileName, tree->siteMode);
      if (DEBUG) fprintf(stderr,"Root sequence scan done ...\n");
    }

    if (refSeqp) {
      if (gOutTrueCM) {
	fprintf(stderr,"main: Will CS_CSTreeCount ...\n");
	CS_CSTreeCount(tree, DcREFSEQ, tree->pCs->rfCsTab, tree->nSites, 
		       tree->pCs->p,  &tree->pCs->nRfCs);
/*  	printCommut(stdout, 1); */
	exit(0);
      }

#if 0
      if (gDoCMStat) {
	/* Here, we apply Shindyalov's criterion to the reference set
	   of sequences */
	for(i=0;i<gNCMRisks;i++) {

	  nPairs = CMstatDetect(tree, DcREFSEQ, mask, numBases, 
				inqBranchNumber(tree, tree->root), 
				&dummyMat, CMTrhld[i], &risk, &rstab, 
				&nTruePairs, 0);
	  fprintf(stdout,"CMStatORI: For threshold %.4lf  : %d pairs detected (%d true).\n",
		  CMTrhld[i], nPairs, nTruePairs);
	}

	/* Ouptut of the boolean matrix associated with the dataset */
	OutputBoolean(oBoolFileName, dummyMat, numBases, 
		      inqBranchNumber(tree, tree->root));
      }
#endif

#if 1
      /* Count number of comutations along tree for reference set (to test) */
      CS_CSTreeCount(tree, DcREFSEQ, tree->pCs->rfCsTab,tree->nSites,
		     tree->pCs->p,  &tree->pCs->nRfCs);
      fprintf(stderr,"RefSeq: Total of %d co-substitutions ... \n", tree->pCs->nRfCs);
#endif

    } /* iRefSeqFileName */

    if (!quiet) {
      PrintVerbose(oLisFich);
      CSParams(oLisFich);
      CS_Infos(oLisFich, tree->pCs);
      RatesParams(oLisFich);
    }

    /* Now we know how many sites will undergo evolution process */
    tmpI = countTrueSites(tree);
    fprintf(stderr,"Total of %d evolving sites\n",tmpI);
    tree->nESites = (double) tmpI;
    
    if (DEBUG) fprintf(stderr,"Will simulate %d evolution(s).\n",numDatasets);
 /*     initLaws(); */
    
    /* Dummy test: prepare everything for simulation */
    /*      fprintf(stderr,"Will dummy.\n"); */

#if 0
#if 0
    SetupMutablepMask(tree,stderr, &maskDtct);
    SetupRateMask(tree, NULL, &maskDtct, 0.01, gMaxRate);
#endif
    dupMask(&maskDtct,tree->mask);

    if (gNoCtctDtct)
      SetupRateMask(tree, NULL, &maskDtct, 0.01, gMaxRate);
    ReadMask(iMaskFileName,&maskDtct);
#endif

    /* ==============================================================
     * If the laws are not input, then compute them
     * i.e: we use pseq-gen simulation facility.
     * ==============================================================
     */
    if (gDoCMStat) {
      for(i=0;i<gNCMRisks;i++) sumCMTrhld[i] = 0.;
      for(i=0;i<gNCMRisks;i++) sumCMRisk[i] = 0.;
    }
    
    /* We accept that tree length is large enough for Gaussian approximation
       of Poisson */
    gDeltaTLen = (gEDTL == 1000.) ? sqrt(tree->refTLen) * gEDTL :
      sqrt(tree->refTLen) * whichEpsilonValue(gEDTL);
    fprintf(stderr,"Tree length: %lf. Allowed variation is: %lf (risk %lf)\n",
	    tree->refTLen, gDeltaTLen, gEDTL);
    fprintf(stderr,"Accepted risk on branch length variation : %lf\n\n",
	    gEDBL);
    
    if (DEBUG) fprintf(stderr,"Will initTreeCount.\n");
    initTreeCount(tree);
    if (DEBUG) fprintf(stderr,"Did initTreeCount.\n"); fflush(stderr);

    
    /* Perform simulation to obtain co-evolution laws --------------- */
    /* Do we equilibrate sequence according to the model before the
       "production" simulation ? */
    /*    if ((gDoEquilibration) && (gEvolveMode == SIM)) { */
    if ((gDoEquilibration) ) {
      fprintf(stderr,"Will perform root equilibration ... \n");
      rootEquilibration(tree, gNEquilSim, scale*branchScale, gEquilSttgy);
    }
    
    /* Adjust scaling for correct tree length simulation -------------- */
    if ((gDoEquilTreeLength) && (gEvolveMode == SIM)) {
      scalingEquilibration(tree, iRootFileName,
			   gNScEquilCycles, gNScEquilSim, 
			   scale, &branchScale, 
			   tree->refTLen, gDeltaTLen,
			   0.01);
    }
    
    /* Attempt to adjust heterogeneity rates to fit "at best" input rates */
    /*    if (( pR->doEquil) && (gEvolveMode == SIM)) { */
    if (( pR->doEquil)) {
      ratesEquilibration(tree, iRootFileName,
			 gNRateEquilCycles, gNRetesEquilSim,
			 gDeltaRate, scale*branchScale,
			 tree->refTLen, gDeltaTLen);

      /* Re-Adjust scaling for correct tree length simulation -------------- */
      if ((gDoEquilTreeLength) && (gEvolveMode == SIM)) {
	scalingEquilibration(tree, iRootFileName,
			     gNScEquilCycles, gNScEquilSim, 
			     scale, &branchScale, 
			     tree->refTLen, gDeltaTLen,
			     0.01);
      }

    }

    
    /* production simulation -------------------------------------- */
    if (iLawsFileName[0] == '\000') {
      if (refSeqp) {
	CS_SSTreeCount(tree, DcREFSEQ, refrates,tree->nSites);
	fprintf(stderr,"Observed number of mutations per site (REFSEQ)\n");
	for (i=0; i<numBases; i++) {
	  if (tree->siteMode[i] != DcCONSTANT)
	    fprintf(stderr,"%6.3lf ",refrates[i]);
	  else 
	    fprintf(stderr," ....  ");
	  if (!((i+1) % 10)) fprintf(stderr,"\n");
	}
	fprintf(stderr,"\n");
      }
      
      meanTLen = 0.;
      stdvTLen = 0.;
      meanCs = 0.;
      stdvCs = 0.;

      if (!quiet) {
	InitProgressBar(numTrees*numDatasets);
	DrawProgressBar();
      }

 /*       gDoPrint = 1; */

      for (aSim=0; aSim<numDatasets; aSim++) {

	/* If we are using random gammaRates, re-assign them */
	SetRates(numBases, gammaShape);

	/* Do we equilibrate sequence according to the model before the
	   "production" simulation ? */
	if (tree->rootMode & DcUSEREQUILIBRATEDROOT) {
	  rootEquilibration(tree, gNEquilSim, scale*branchScale, gEquilSttgy);
	}
    
	/* Perform a simulation according to some conditions */
	doEvolveSequences(tree, iRootFileName, tree->pCs->p, 
			  scale*branchScale, 
			  gEDBL, gNBMaxTries, 
			  &curTLen, gDeltaTLen, gDeltaRate, gNTMaxTries);

	meanTLen += curTLen;
	stdvTLen += curTLen *  curTLen;
	
	if (gEvolveMode != SIM) {
	  CS_Usage(tree->pCs);
	}
	
	/*  	  if (0 & gShowHomol) TipsIdentity(stderr, tree); */
	/* We might want to store the sequences for further analysis */
	WriteSequences(oSeqFich, tree, gOutNodes);
	
/*  	GetEntropies(tree); */
/*  	storeTreeLengths(tree, DcSIMSEQ); */

	if (CS_detect) {
	  CS_CSTreeCount(tree, DcSIMSEQ, tree->pCs->csTab, tree->nSites,
			 tree->pCs->p, &tree->pCs->nCs);
/*  	  fprintf(stderr,"SimSeq: Total of %d co-substitutions ... \n", tree->pCs->nCs); */
	meanCs += (double) tree->pCs->nCs; 
	stdvCs += (double) tree->pCs->nCs * (double) tree->pCs->nCs;
	
	  /* Build laws of number of co-substitutions among simulations */
	  CS_lawsBuild(tree->pCs, tree->pCs->csTab);

	  /* Compare reference and simulated number of co-substitutions */
	  CS_CSCmp(tree->nSites, tree->pCs->rfCsTab, tree->pCs->csTab, 
		    tree->pCs->rsTab);
	}
	
	if (!quiet)
	  ProgressBar();
      }  /* numDatasets loop */
      
/*        fprintf(stderr,"%d rate failures upon %d simulations\n",nRatesFail,numDatasets); */
      if ( numDatasets > 1)
	stdvTLen = (stdvTLen - (meanTLen * meanTLen) / (double) numDatasets) / 
	  ((double) numDatasets - 1);
      else
	stdvTLen = (stdvTLen - (meanTLen * meanTLen) / (double) numDatasets) / 
	  ((double) numDatasets);
      meanTLen /= (double) numDatasets;
      fprintf(stderr,"mean Tree Len %.2lf stdev %.2lf (target %.2lf)\n",
	      meanTLen,sqrt(stdvTLen),tree->refTLen);
      
      if ( numDatasets > 1)
	stdvCs = (stdvCs - (meanCs * meanCs) / (double) numDatasets) / 
	  ((double) numDatasets - 1);
      else
	stdvCs = (stdvCs - (meanCs * meanCs) / (double) numDatasets) / 
	  ((double) numDatasets);
      meanCs /= (double) numDatasets;
      fprintf(stderr,"mean number of CS %.2lf stdev %.2lf (reference is %d)\n",
	      meanCs,sqrt(stdvCs), tree->pCs->nRfCs);
      
   /*     setGlobLaw(&meanCmt, &varCmt); */
      CS_OutputLaws(oLawsFileName, tree->pCs);
    } else { /* If we just read laws from disk */
      CS_InputLaws(iLawsFileName, tree->pCs);
    }

    /* ==============================================================
     * Now: perform detection
     * ==============================================================
     */

    if (! CS_detect) goto LNEXT;
    fprintf(stderr,"Detection of significant co-subsituting pairs: \n");
    fprintf(stderr,"target thresholds:");
    for(i=0;i<gNRisks;i++) {
      fprintf(stderr," %lf",gRisks[i]);
    }
    fprintf(stderr,"\n");

    CS_initThresholds(tree->pCs);
    for(i=0;i<gNRisks;i++) {
      risk = gRisks[i];
      CS_doDetect(tree->pCs, risk, tree->mask, numDatasets, verbose); 
    }
    
  LNEXT:;
    
    DisposeTree(tree);
    treeNo++;
    fprintf(stderr,"Total of %d branch evolutions performed for %d used\n",gNTries, gNEvolve);
  } while (IsTreeAvail(tree_fv));
  
  FreeTree(tree);
  /*  Free_CM();  */                      /* M D 1997  */
  if (oSeqFich != NULL) fclose(oSeqFich);
  
  totalSecs = (double)(clock() - totalStart) / CLOCKS_PER_SEC;
  if (!quiet) {
    fprintf(stderr, "Time taken: %G seconds\n", totalSecs);
    if (verboseMemory)
      fprintf(stderr, "Total memory used: %ld\n", totalMem);
  }
}
